/* Compact floating-point encoding, look at arguments specifying decode table.
   Written by Radford M. Neal, 2015. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "cflt.h"

int cflt_args (char **argv)
{
  int bits, bits2, epos, i, j, k, n, m, cant, unique;
  int t_used, m_used, e_used;
  double flt;

  static char ddd[11] = "dddddddddd";
  static char zzz[11] = "0000000000";

  static double pow10[20] = 
  {  1e0,  1e1,  1e2,  1e3,  1e4,  1e5,  1e6,  1e7,  1e8,  1e9,
    1e10, 1e11, 1e12, 1e13, 1e14, 1e15, 1e16, 1e17, 1e18, 1e19
  };

  if (!argv[1]) return -1;
  bits = atoi(argv[1]);
# if CFLT_2_MASKS | CFLT_RECODE
    if (!argv[2]) return -1;
    bits2 = atoi(argv[2]);
    if (!argv[3]) return -1;
    epos = atoi(argv[3]);
    printf("\nIndex is low %d mantissa bits + %d exponent bits offset %d\n\n",
            bits, bits2, epos);
    cflt_start(bits,bits2,epos);
    i = 4;
# else
    bits2 = 0;
    epos = 0;
    printf("\nIndex is low %d bits\n\n",bits);
    cflt_start(bits);
    i = 2;
# endif

  for ( ; argv[i]; i++)
  {
    char *p = argv[i];
    char c;

    if (strcmp(p,"NA") == 0)
    { printf("Using NA\n");
      if (!cflt_use_parts(0x7fffffff,1954)) /* low part matches NA value in R */
      { printf("Can't compress NA\n");
      }
      continue;
    }

    n = strtol (p, &p, 10);
    if (!*p) return -1;
    c = *p++;
    m = strtol (p, &p, 10);
    if (*p) return -1;
  
    switch (c)
    { 
      case '/':
      { if (n < 1 || m < 1) return -1;
        printf("Using numbers that are ratios up to %d/%d\n",n,m);
        cant = 0;
        for (j = 0; j <= n; j++)
        { for (k = 1; k <= m; k++)
          { flt = (double)j / k;
            if (!cflt_use_value(flt))
            { if (cant < 3) printf("  Can't compress %d/%d = %f\n",j,k,flt);
              cant += 1;
            }
          }
        }
        break;
      }

      case '.':
      { if (n <= -10 || n >= 10 || m <= -10 || m >= 10) return -1;
        if (n < 0 && m < 0) return -1;
        printf("Using numbers of the form %s%s.%s%s\n",
                n < 0 ? "" : ddd+10-n, m >= 0 ? "" : zzz+10+m,
                n >= 0 ? "" : zzz+10+n, m < 0 ? "" : ddd+10-m);
        cant = 0;
        for (j = 0; j < pow10 [ n<0 ? m : m<0 ? n : n+m ]; j++)
        { flt = m >= 0 ? (double)j / pow10 [ n>=0 ? m : m-n ]
                       : (double)j * pow10 [ -m ];
          if (!cflt_use_value(flt))
          { if (cant < 3) 
            { union { double d; uint64_t v; } u;
              u.d = flt;
              printf("  Can't compress %.*f (%016llx)\n",
                        m<0 ? 0 : n>=0 ? m : m-n, flt,
                        (long long int) u.v);
            }
            cant += 1;
          }
        }
        break;
      }

      default: return -1;
    }

    if (cant > 3) printf("  ...\n  Total of %d can't be compressed\n",cant);
  }

  cflt_number_used (&t_used, &m_used, &e_used);

  printf ("\nUsed %d table entries out of %d\n", t_used, cflt_size);

# if CFLT_2_MASKS | CFLT_RECODE
    printf("Used %d out of %d mantissa values, %d out of %d exponent values\n",
            m_used, 1<<bits, e_used, 1<<bits2);
# endif

# if CFLT_RECODE
    cflt_recode();
    printf ("Recoded to %d entries (%d x %d)\n", 
             cflt_size, (1 << cflt_bits2), cflt_ncodes);
# endif

  unique = cflt_indirect();

  printf ("There are %d unique values for the low part of the mantissa\n", 
           unique);

# if CFLT_INDIRECT
#   if CFLT_RECODE
      printf("Bytes of memory required for tables: %d + %d + %d = %d\n\n", 
       (int) ((1 << cflt_bits) * sizeof *cflt_rtable),
       (int) (cflt_size*sizeof *cflt_itable), 
       (int) (unique*sizeof *cflt_table),
       (int) ((1 << cflt_bits) * sizeof *cflt_rtable) +
       (int) (cflt_size*sizeof *cflt_itable) +
       (int) (unique*sizeof *cflt_table));
#   else
      printf("Bytes of memory required for tables: %d + %d = %d\n\n", 
       (int) (cflt_size*sizeof *cflt_itable),
       (int) (unique*sizeof *cflt_table),
       (int) (cflt_size*sizeof *cflt_itable + unique*sizeof *cflt_table));
#   endif
# else
#   if CFLT_RECODE
      printf("Bytes of memory required for tables: %d + %d = %d\n\n", 
             (int) ((1 << cflt_bits) * sizeof *cflt_rtable),
             (int) (cflt_size * sizeof *cflt_table),
             (int) ((1 << cflt_bits) * sizeof *cflt_rtable) +
             (int) (cflt_size * sizeof *cflt_table));
#   else
      printf("Bytes of memory required for table: %d\n\n", 
             (int) (cflt_size * sizeof *cflt_table));
#   endif
# endif

  return 0;
}
