/* Program to use with "top" to check when physical memory is allocated. */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <time.h>
#include <unistd.h>
#include <sys/mman.h>

int main (void)
{
  char *space;
  int len;
  char c;
  clock_t start;
  int i, j, k, p, v;
  int pgsize;

  pgsize = getpagesize();
  printf("PAGESIZE: %d\n",pgsize);

  len = 0;
  for (;;)
  { printf("?");
    if (scanf(" %c",&c) !=  1) break;
    if (c == 'a')
    { if (scanf("%d",&i) != 1) goto error;
      start = clock();
      space = malloc(i);
      if (space==0) { printf("malloc of %d failed\n",i); return 1; }
      len = i;
      printf("allocated %d bytes in %.3f seconds\n",i,
               (double)(clock()-start)/CLOCKS_PER_SEC);
    }
    else if (c == 's')
    { if (scanf("%d %d",&i,&j)!=2) goto error;
      if (i < 0 || i >= len || j < i || j >= len) 
      { printf("indexes %d, %d out of range\n",i,j);
      }
      else
      { start = clock();
        for (k = i; k <=j; k++) space[k] = 123;
      }
      printf("stored 123 in %d, %d in %.3f seconds\n",i,j,
             (double)(clock()-start)/CLOCKS_PER_SEC);
    }
    else if (c == 'f')
    { if (scanf("%d %d",&i,&j)!=2) goto error;
      if (i < 0 || i >= len || j < i || j >= len) 
      { printf("indexes %d, %d out of range\n",i,j);
      }
      else
      { v = 0;
        start = clock();
        for (k = i; k <= j; k++) v |= space[k];
        printf("fetched from %d, %d, ors to %d in %.3f seconds\n",i,j,v,
               (double)(clock()-start)/CLOCKS_PER_SEC);
      }
    }
    else if (c == 'd')
    { if (scanf("%d %d",&i,&j)!=2) goto error;
      if (i < 0 || i >= len || j < i || j >= len) 
      { printf("indexes %d, %d out of range\n",i,j);
      }
      else
      { uintptr_t st = (uintptr_t) (space+i);
        uintptr_t en = (uintptr_t) (space+j);
        uintptr_t st2, en2;
        st2 = (st + pgsize-1) & ~(uintptr_t)(pgsize-1);
        en2 = en & ~(uintptr_t)(pgsize-1);
        printf ("range %08lx/%08lx to %08lx/%08lx\n", 
                 (unsigned long)st, (unsigned long)st2, 
                 (unsigned long)en, (unsigned long)en2);
        if (st2 < en2)
        { if (mmap ((void *)st2, (size_t)(en2-st2), PROT_WRITE, 
                     MAP_PRIVATE | MAP_FIXED | MAP_ANONYMOUS, -1, 0)
              == MAP_FAILED)
          { printf("failed\n");
          }
          else
          { printf("succeeded\n");
          }
        }
        else
        { printf("empty\n");
        }
      }
    }
    continue;

  error:
    printf("invalid command\n");
  }

  return 0;
}
